library spark;

import 'dart:html' as html;
import 'package:sparkflow/sparkflow.dart';
import 'package:hub/hubclient.dart';
import 'package:spark_elements/elements.dart';

class Spark{
  static registerComponents(){
      Component.registerComponents();
      Elements.registerComponents();
      SparkRegistry.register('spark','domcards',DOMCards.create);
      SparkRegistry.register('spark','eventfilter',EventFilter.create);
      SparkRegistry.register('spark','eventreactor',EventReactor.create);
      SparkRegistry.register('spark','filteredeventreactor',FilteredEventReactor.create);
  }

  static isEvent(n) => n is html.Event;
}

class FilteredEventReactor extends Component{

  static create() => new FilteredEventReactor();

  FilteredEventReactor(){

  }
}


class EventReactor extends Component{

  static create() => new EventReactor();

  EventReactor(){
      
      Function runner;

      this.enableSubnet();
      this.createSpace('evin');

      this.makeInport('evin:events',meta:{'desc':"stream of events"});
      this.makeInport('evin:fn',meta:{"desc":"a function to apply all events with "});
      
      this.port('static:option').bindPort(this.port('evin:fn'));
  
      this.port('evin:events').forceCondition(Spark.isEvent);
      this.port('evin:fn').forceCondition(Valids.isFunction);

      this.port('evin:events').pause();

      this.port('evin:fn').tap((n){
         runner = n.data;
         this.port('evin:events').resume();
      });
      

  }

}

class DOMCards extends Element{

  static create() => new DOMCards();

  DOMCards(){
  
    this.mutate((shell){

        MapDecorator eventcards = MapDecorator.create();

        var notifyEvent = (packet){
          if(eventcards.has(packet.data)) return null;
          var fn = (e){
            var pack = shell.port('events:events').createDataPacket(e);
            pack.update('eventType',packet.data);
            return shell.port('events:events').send(pack);
          };

          eventcards.add(packet.data,fn);
          shell.elem.addEventListener(packet.data,fn);
        };

        var unNotifyEvent = (packet){
          if(!eventcards.has(packet.data)) return null;
          var fn = eventcards.destroy(packet.data);
          shell.elem.addEventListener(packet.data,fn);
        };

        shell.meta('desc','provides a self contained dom element');

        //adds a new event group space for domelement events
        shell.createSpace('events');
        //takes in a string name of an event to be notified about
        shell.makeInport('events:sub');
        //takes in a string id of an event to stop notifying about
        shell.makeInport('events:unsub');
        //sends out event packets that occur with the dom element which it is listening to
        shell.makeOutport('events:events');

        shell.port('events:sub').pause();
        shell.port('events:unsub').pause();
      
        shell.port('events:sub').forceCondition(Valids.isString);
        shell.port('events:unsub').forceCondition(Valids.isString);

        shell.port('events:sub').tap(notifyEvent);
        shell.port('events:unsub').tap(unNotifyEvent);

        shell.port('events:sub').tap(notifyEvent);
        shell.port('events:unsub').tap(unNotifyEvent);

        shell.port('in:elem').tap((pack){
          
          shell.port('events:sub').resume();
          shell.port('events:unsub').resume();
    /**/
    /*      shell.port('events:sub').send('error');*/
    /*      shell.port('events:sub').send('focus');*/
    /*      shell.port('events:sub').send('drag');*/
    /*      shell.port('events:sub').send('click');*/
    /*      shell.port('events:sub').send('select');*/
          
        });

    });

  }


}

class EventFilter extends Component{
  var allowed = "all";
  bool reverse = false;

  static create() => new EventFilter();

  EventFilter(){

    this.createSpace('ev');
    this.createSpace('state');
    this.createSpace('type');

    this.makeInport('type:allowed',
        meta:{'description':'takes a string or list of allowed event names' });
    this.makeInport('ev:events',
        meta:{'description':"the port where all events comes in"});
    this.makeInport('state:invert',
        meta:{'description':"takes a bool operation that inverts the operation"});

    this.makeOutport('state:pass');
    this.makeOutport('state:fail');
    
    this.port('state:invert').forceCondition(Valids.isBool);
    this.port('ev:events').forceCondition(Spark.isEvent);

    this.port('state:invert').tap((n){
      this.reverse = n.data;
    });

    this.port('type:allowed').forceCondition((n){
      return (Valids.isString(n) || Valids.isList(n) ? true : false);
    });

    this.port('type:allowed').tap((n){
      this.allowed = n.data;
    });
  
    this.port('ev:events').tap(this.filter);

  }

  void _handle(fn,fe,n){
      var type = n.has('eventType') ? n.get('eventType') : n.data.type;

      if(Valids.isString(this.allowed)){
      
        Funcs.when(Valids.match(this.allowed,"all"),(){
           return fn(n);
        });

        Funcs.when(Valids.match(this.allowed,"none"),(){
           return fe(n);
        });

        Funcs.when(Valids.match(type,this.allowed),(){
           return fn(n);
        },(){
          return fe(n);
        });
  
        return null;
      }


      if(Valids.isList(this.allowed))
        return Funcs.when(this.allowed.contains(type),(){
          return fn(n);
        },(){
          return fe(n);
        });

      return fe;
  }

  void filter(m){

    Funcs.when(this.reverse,(){

       return this._handle((n){
          return this.sendFailure(n);
       },(n){
          return this.port('state:pass').send(n);
       },m);

    },(){
        
       return this._handle((n){
          return this.port('state:pass').send(n);
       },(n){
          return this.sendFailure(n);
       },m);

    });


  }

  void sendFailure(n){
    var fail = this.port('state:fail');
    if(fail.hasSubscribers) return fail.send(n);
    return null;
  }

}
